/*
 * RubyMalaga
 * Copyright  2002, 2005 Michael Piotrowski <mxp@dynalabs.de>
 *
 * All rights reserved.  You can redistribute and/or modify this library
 * under the same terms as Ruby.
 *
 */

#include <ruby.h>
#include <malaga.h>

#include <stdio.h>

/* $Format: "static char version[] = \"@(#) $ProjectVersion$\";"$ */
static char version[] = "@(#) 0.1";
static char ident[]   = "@(#) $Id: malaga.c 1.1 Mon, 02 May 2005 18:59:35 +0200 mxp $";

VALUE cMalagaGrammar;
VALUE cMalagaAnalyses;
VALUE cMalagaAnalysis;
VALUE eMalagaError;


static VALUE rm_option (VALUE self, VALUE option_key, VALUE option_val);


static VALUE malaga_to_ruby (value_t value)
{
   VALUE ruby_val;
   VALUE ruby_array;
   VALUE ruby_hash;

   switch (get_value_type (value))
   {
      case SYMBOL_SYMBOL:
      {
	 ruby_val = ID2SYM(
	    rb_intern((char *) get_symbol_name(value_to_symbol (value))));
	 break;
      }
  
      case STRING_SYMBOL:
      {
	 ruby_val = rb_str_new2((char *) value_to_string(value));

	 break;
      }
      
      case LIST_SYMBOL:
      {
         value_t item;

	 ruby_array = rb_ary_new();

	 for (item = get_first_item (value);
	      item != NULL;
              item = get_next_item (value, item))
	 {
	    rb_ary_push(ruby_array, malaga_to_ruby(item));
	 }

	 ruby_val = ruby_array;

	 break;
      }

      case RECORD_SYMBOL:
      {
         value_t item;

	 ruby_hash = rb_hash_new();

	 for (item = get_first_item (value);
              item != NULL;
              item = get_next_item (value, item))
	 {
	    symbol_t attrib_symbol = value_to_symbol(item);
	    string_t attrib_name   = get_symbol_name(attrib_symbol);

	    rb_hash_aset(ruby_hash, rb_str_new2((char *) attrib_name),
			 malaga_to_ruby(get_attribute(value, attrib_symbol)));

	 }
	 ruby_val = ruby_hash;
	 break;
      }
      
      case NUMBER_SYMBOL:
      {
	 ruby_val = rb_float_new(value_to_double(value));
	 break;
      }
   } /* end of switch */

   return ruby_val;
}

static VALUE tt_yield(VALUE arg1, VALUE arg2)
{
   VALUE key, value;

   key = rb_ary_entry(arg1, 0);
   value = rb_ary_entry(arg1, 1);
   /* do what you want with key and value */

   printf("%s // %s\n", STR2CSTR(key), STR2CSTR(value));

   return Qnil;
}

static VALUE handle_activation (VALUE grammar)
{
   if (rb_cv_get(rb_obj_class(grammar), "@@active_grammar") != grammar)
   {
      if (rb_cv_get(rb_obj_class(grammar), "@@active_grammar"))
      {
	 terminate_libmalaga();
      }

      init_libmalaga(STR2CSTR(rb_iv_get(grammar, "@grammar")));

      if (malaga_error)
	 rb_raise(eMalagaError, malaga_error);


      rb_iterate(rb_each, rb_iv_get(grammar, "@options"),
		 tt_yield, 0);


      rb_cv_set(rb_obj_class(grammar), "@@active_grammar", grammar);
   }
   
   return Qtrue;
}

/* ********************************************************************** */

/*
 * Creates a MalagaAnalysis object.  Only used internally.
 *
 */

static VALUE rma_init (VALUE self, VALUE surface, value_t malaga_val)
{
   Check_Type(surface, T_STRING);

   rb_iv_set(self, "@surface", surface);
   rb_iv_set(self, "@category", malaga_to_ruby(malaga_val));

   return self;
}

/*
 * Returns the text that resulted in this analysis (the +text+
 * argument to MalagaGrammar#analyze, MalagaGrammar#analyze_m or
 * MalagaGrammar#analyze_s).
 *
 */

static VALUE rma_surface (VALUE self)
{
   return rb_iv_get(self, "@surface");
}

/*
 * Returns the category, i.e., the analysis proper.  This is the
 * Malaga analysis converted to Ruby classes; for example, if the
 * analysis returns a record, you'll get a Ruby Hash object.
 * 
 * Malaga data types are mapped to Ruby classes as follows:
 * 
 *   Malaga type           | Ruby class
 *   ----------------------+------------------------
 *   Symbol                | Symbol
 *   String                | String
 *   Floating point number | Float
 *   List                  | Array
 *   Record                | Hash
 * 
 * 
 * For example, the following Malaga analysis
 * 
 *   1: [clause_type: declarative,
 *       predicate: "ist",
 *       arguments: <[surf: "Gero",
 *                    function: subject],
 *                   [surf: "in der Stadt",
 *                    function: local]>]
 * 
 * results in the following Ruby structure (as printed by pp):
 * 
 *   {"arguments"=>
 *     [{"surf"=>"Gero", "function"=>:subject},
 *      {"surf"=>"in der Stadt", "function"=>:local}],
 *    "predicate"=>"ist",
 *    "clause_type"=>:declarative}
 * 
 */

static VALUE rma_category (VALUE self)
{
   return rb_iv_get(self, "@category");
}

/*
 * As a shortcut, you can use
 * 
 *   a['CG']
 * 
 * instead of
 * 
 *   a.category['CG']
 *
 */

static VALUE rma_cat_direct (VALUE self, VALUE key)
{
   return rb_funcall(rb_iv_get(self, "@category"),
		     rb_intern("[]"), 1, key);
}

/* internal */

static VALUE rma_cat_direct_s (VALUE self)
{
   return rb_funcall(rb_iv_get(self, "@category"),
		     rb_intern("to_s"), 0);
}



static VALUE tt_iter_m(VALUE *tmp)
{
   return rb_funcall2(tmp[2], tmp[3], (int)tmp[0], (VALUE *)tmp[1]);
}
static VALUE tt_yield_m(VALUE obj, VALUE res)
{
   rb_ary_push(res, rb_yield(obj));
   return Qnil;
}

/* internal */

static VALUE rma_missing (int argc, VALUE *argv, VALUE self)
{
   ID id;

   if (argc == 0 || !SYMBOL_P(argv[0]))
   {
      rb_raise(rb_eArgError, "no id given");
   }

   id = SYM2ID(argv[0]);

/* */
   if (rb_block_given_p())
   {
      VALUE res, tmp[4];

      /* printf("XXX block given\n"); */

      tmp[0] = (VALUE)argc - 1;
      tmp[1] = (VALUE)argv + 1;
      tmp[2] = rb_iv_get(self, "@category");
      tmp[3] = id;
      res = rb_ary_new();
      rb_iterate(tt_iter_m, (VALUE)tmp, tt_yield_m, res);
      return res;
   }





/* */

/* better? doesn't work for methods with no arguments */
/*       return rb_funcall2(rb_iv_get(self, "@category"), */
/* 			 id, argc, argv); */

   return rb_funcall(rb_iv_get(self, "@category"),
		     id, argc - 1, argv + 1);
}

/* ********************************************************************** */

/*
 * call-seq: Malaga.new(project_filename) -> Malaga
 *
 * Creates a Malaga grammar object; +project_filename+ is the filename
 * of the Malaga project file.
 *
 * Example:
 *
 *   g = Malaga.new('/usr/local/lib/malaga/imm/imm.pro')
 * 
 */

static VALUE rm_init(VALUE self, VALUE project_filename)
{
   Check_Type(project_filename, T_STRING);

   rb_iv_set(self, "@grammar", project_filename);
   rb_iv_set(self, "@options", rb_hash_new());
   return self;
}

/*
 * Returns the name of the project file name used by the grammar object.
 *
 */

static VALUE rm_grammar (VALUE self)
{
   return rb_iv_get(self, "@grammar");
}

/*
 * call-seq: info -> string
 *
 * Returns a string with the informative message specified in the
 * "info" lines of the project file.
 */

static VALUE rm_info (VALUE self)
{
   handle_activation(self);

   return rb_str_new2(get_info());
}

/*
 * call-seq: option(option_key, option_value)
 *
 * Allows to set options for the loaded grammar.  This allows to
 * override options specified in the project file.  The possible
 * options are described in the Malaga manual.
 *
 * Returns true on success, otherwise raises a MalagaError exception.
 *
 * Example:
 *
 *  g.option("sort-records", "alphabetic")
 */

static VALUE rm_option (VALUE self, VALUE option_key, VALUE option_val)
{
   VALUE option_string;

   handle_activation(self);

   option_string = rb_str_dup(option_key);
   option_string = rb_str_concat(option_string, rb_str_new2(" "));
   option_string = rb_str_concat(option_string, option_val);

   set_option((string_t) STR2CSTR(option_string));
   if (malaga_error)
      rb_raise(eMalagaError, malaga_error);

   rb_hash_aset(rb_iv_get(self, "@options"), option_key, option_val);

   return Qtrue;
}

/*
 * call-seq: options -> hash
 *
 * Returns a hash containing the current values of options; it only
 * contains options that were set using #option, options set in the
 * project file don't appear here.
 *
 */

static VALUE rm_options (VALUE self)
{
   return rb_iv_get(self, "@options");
}

static VALUE rm_analyze (VALUE self, VALUE text, int analysis_type)
{
   value_t result;
   VALUE ruby_result;

   handle_activation(self);

/*    ruby_result = rb_ary_new(); */
   ruby_result = rb_funcall(cMalagaAnalyses, rb_intern("new"), 0);

   analyse_item((string_t) STR2CSTR(text), analysis_type);

   if (malaga_error)
      rb_raise(eMalagaError, malaga_error);

   if ((result = first_analysis_result()) != NULL)
   {
/*       rb_ary_push(ruby_result, malaga_to_ruby(result)); */
      rb_ary_push(ruby_result,
		  rb_funcall(cMalagaAnalysis, rb_intern("new"), 2,
			     text, result));

      while ((result = next_analysis_result()) != NULL)
      {
/* 	 rb_ary_push(ruby_result, malaga_to_ruby(result)); */
	 rb_ary_push(ruby_result,
		     rb_funcall(cMalagaAnalysis, rb_intern("new"), 2,
				text, result));
      }
   }

   return ruby_result;
}

/*
 * call-seq: analyze_m(text) -> MalagaAnalyses
 *
 * Analyzes the string given in +text+ morphologically.  Returns a
 * MalagaAnalyses object.
 *
 */

static VALUE rm_analyze_mor (VALUE self, VALUE text)
{
   return rm_analyze(self, text, MORPHOLOGY);
}

/*
 * call-seq: analyze_s(text) -> MalagaAnalyses
 *
 * Analyzes the string given in +text+ syntactically.  Returns a
 * MalagaAnalyses object.
 *
 */

static VALUE rm_analyze_syn (VALUE self, VALUE text)
{
   return rm_analyze(self, text, SYNTAX);
}

/* ********************************************************************** */

void Init_malaga ()
{
   cMalagaGrammar = rb_define_class("MalagaGrammar", rb_cObject);
   eMalagaError = rb_define_class("MalagaError", rb_eException);

   rb_define_class_variable(cMalagaGrammar, "@@active_grammar", Qnil);

   rb_define_method(cMalagaGrammar, "initialize", rm_init, 1);
   rb_define_method(cMalagaGrammar, "grammar", rm_grammar, 0);
   rb_define_method(cMalagaGrammar, "info", rm_info, 0);
   rb_define_method(cMalagaGrammar, "option", rm_option, 2);
   rb_define_method(cMalagaGrammar, "options", rm_options, 0);
   rb_define_method(cMalagaGrammar, "analyze_m", rm_analyze_mor, 1);
   rb_define_method(cMalagaGrammar, "analyze_s", rm_analyze_syn, 1);

   rb_define_alias(cMalagaGrammar, "analyze", "analyze_m");

   cMalagaAnalyses = rb_define_class("MalagaAnalyses", rb_cArray);

   cMalagaAnalysis = rb_define_class("MalagaAnalysis", rb_cObject);
   rb_define_method(cMalagaAnalysis, "initialize", rma_init, 2);
   rb_define_method(cMalagaAnalysis, "surface", rma_surface, 0);
   rb_define_method(cMalagaAnalysis, "category", rma_category, 0);

   rb_define_method(cMalagaAnalysis, "[]", rma_cat_direct, 1);
   rb_define_method(cMalagaAnalysis, "to_s", rma_cat_direct_s, 0);

   rb_define_method(cMalagaAnalysis, "method_missing", rma_missing, -1);
}
